# This script was used to scrub the harness of old code
#   First a set of sed patterns found all the harness classes
#   that are in used directly.  They were moved out of place using,
#   this script, then those that were used indirectly were moved
#   back in place.
import glob
import re
import os

with open("keep", "r") as fd:
    lines = fd.readlines()
    lines = [i.strip() for i in lines]
    keep = set(lines)

for fl in glob.iglob('harness/**/*.java', recursive=True):
    m = re.match('.*(jpype.*).java', fl)
    name = m.group(1).replace('/', '.')
    if name in keep:
        continue
    m = re.match('(.*)/[^/]*.java', fl)
    os.makedirs(os.path.join("old", m.group(1)), exist_ok=True)
    os.rename(fl, os.path.join("old", fl))



# Utility used to alter a Java jar file to date it in the future.
# This was needed to intentionally trigger a version error.
import jpype
import jpype.imports
jpype.startJVM()

from java.util.jar import JarOutputStream
from java.util.jar import JarInputStream
from java.util.zip import CRC32
from java.io import File
from java.io import FileInputStream
from java.io import FileOutputStream

jar = JarInputStream(FileInputStream(File("build/lib/org.jpype.jar")))
manifest = jar.getManifest()
target = JarOutputStream(FileOutputStream(File("build/lib/org.jpype2.jar")), manifest)


while 1:
    entry = jar.getNextEntry()
    if not entry:
        break
    out = []
    l3 = 512
    while 1:
        bt = jpype.JArray(jpype.JByte)(l3)
        l = jar.read(bt, 0, l3)
        if l == -1:
            break
        out.append((l, bt))

    if out:
        out[0][1][7] = 57

        crc = CRC32()
        for v in out:
            crc.update(v[1], 0, v[0])

        entry.setCrc(crc.getValue())
        entry.setCompressedSize(-1)

    target.putNextEntry(entry)
    for v in out:
        target.write(v[1], 0, v[0])
    target.closeEntry()
target.close()



import jpype
from typing import Union


class Iterable:
    """ Customized interface for a container which can be iterated.

    JPype wraps ``java.lang.Iterable`` as the Python iterator
    interface.
    """

    def __iter__(self):
        """ Iterate over the members on this collect. """
        ...


class Collection:
    """ Customized interface representing a collection of items.

    JPype wraps ``java.util.Collection`` as a Python collection.
    """

    def __len__(self) -> int:
        """ Get the length of this collection.

        Use ``len(collection)`` to find the number of items in this
        collection.

        """
        ...

    def __delitem__(self, item):
        """ Collections do not support remove by index. """
        ...

    def __contains__(self, item) -> bool:
        """ Check if this collection contains this item.

        Use ``item in collection`` to check if the item is 
        present.

        Args:
           item: is the item to check for.  This must be a Java
           object or an object which can be automatically converted
           such as a string.

        Returns:
           bool: True if the item is in the collection.
        """
        ...


class List(Collection):
    """ Customized container holding items in a specified order.

    JPype customizes ``java.lang.List`` to be equivalent to a Python list.
    Java list fulfill the contract for ``collection.abc.MutableSequence``.
    """

    def __getitem__(self, ndx):
        """ Access an item or set of items.

        Use ``list[idx]`` to access a single item or ``list[i0:i1]`` to obtain
        a slice of the list.  Slices are views thus changing the view will
        alter the original list.  Slice stepping is not supported for Java
        lists.
        """
        ...

    def __setitem__(self, index: Union[int, slice], value):
        """ Set an item on the list.

        Use ``list[idx]=value`` to set a value on the list or 
        ``list[i0:i1] = values`` to replace a section of a list with
        another list of values.
        """
        ...

    def __delitem__(self, idx: Union[int, slice]):
        """ Delete an item by index.

        Use ``del list[idx]`` to remove ont itme from the list or
        ``del list[i0:i1]`` to remove a section of the list.
        """
        ...

    def __reversed__(self):
        """ Obtain an iterator that walks the list in reverse order.

        Use ``reversed(list)`` to traverse a list backwards.
        """
        ...

    def index(self, obj) -> int:
        """ Find the index that an item appears.

        Args:
            obj: A Java object or Python object which automatically
                converts to Java.

        Returns:
            int: The index where the item first appears in the list.

        Raises:
           ValueError: If the item is not on the list.
        """
        ...

    def count(self, obj):
        """ Count the number of times an object appears in a list.

        Args:
            obj: A Java object or Python object which automatically
                converts to Java.

        Returns:
            int: The number of times this object appears.
        """
        ...

    def insert(self, idx: int, obj):
        """ Insert an object at a specific position.

        Args:
            idx: The index to insert the item in front of.
            obj: The object to insert.

        Raises:
            TypeError: If the object cannot be converted to Java.
        """
        ...

    def append(self, obj):
        """ Append an object to the list.

        Args:
            obj: The object to insert.

        Raises:
            TypeError: If the object cannot be converted to Java.
        """
        ...

    def reverse(self):
        """ Reverse the order of the list in place.

        This is equivalent to ``java.util.Collections.reverse(list)``.
        """

    def extend(self, lst):
        """ Extends a list by adding a set of elements to the end.

        Args:
           lst: A Sequence holding items to be appended.

        Raises:
           TypeError: If the list to be added cannot be converted to Java.
        """
        ...

    def pop(self, idx=-1):
        """ Remove an item from the list.

        Args:
           idx (int, optional): Position to remove the item from.  If not
             specified the item is removed from the end of the list.

        Returns:
           The item or raises if index is outside of the list.
        """
        ...

    def __iadd__(self, obj):
        """ Add an items to the end of the list.

        Use ``list += obj`` to append one item.  This is simply an alias
        for add.
        """
        ...

    def __add__(self, obj):
        """ Combine two lists.

        Use ``list + seq`` to create a new list with additional members.
        This is only supported if the list can be cloned.
        """
        ...

    def remove(self, obj):
        """ Remove an item from the list by finding the first
        instance that matches.

        This overrides the Java method to provide the Python remove.
        Use ``lst.remove_`` to obtain the Java remove method.

        Args:
           obj: Must be a Java object or Python object that can 
             convert to Java automatically.

        Raises:
            ValueError: If the item is not present on the list.
        """
        ...


class Map:
    """ Customized container holding pairs of items like a dictionary.

    JPype customizes ``java.lang.List`` to be equivalent to a Python list.
    Java maps fulfill the contract for ``collection.abc.Mapping``.
    """

    def __len__(self):
        """ Get the number of items in this map.

        Use ``len(map)`` to get the number of items in the map.
        """
        ...

    def __iter__(self):
        """ Iterate the keys of the map.
        """
        ...

    def __delitem__(self, i):
        """ Remove an item by its key.

        Raises:
           TypeError: If the key cannot be converted to Java.
        """
        ...

    def __getitem__(self, ndx):
        """ Get a value by its key.

        Use ``map[key]`` to get the value associate with a key. 

        Raises:
           KeyError: If the key is not found in the map or the key
             cannot be converted to Java.
        """
        ...

    def __setitem__(self, key, value):
        """ Set a value associated with a key..

        Use ``map[key]=value`` to set the value associate with a key. 

        Raises:
           TypeError: If the key or value cannot be converted to Java.
        """
        ...

    def items(self):
        """ Get a list of entries in the map.

        The map entries are customized to appear as tuples with two 
        items.  Maps can traversed as key value pairs using ``map.items()``
        """
        ...

    def keys(self) -> list:
        """ Get a list of keys for this map.

        Use ``map.keySet()`` to obtain the keys as Java views them.

        Returns:
           list: A Python list holding all of the items.
        """
        ...

    def __contains__(self, item):
        """ Check if a key is in the map.

        Use ``item in map`` to verify if the map contains the item.
        This will return true whether on not the associated value
        is an object or None.

        Returns:
          True is the key is found.
        """
        ...


class Set(object):
    """ Customized Java Sets.

    Java sets only provide the ability to delete items.
    """
    ...


class Iterator:
    """ Customized Java Iterator.

    Java iterators act just like Python iterators for the 
    purposed of list comprehensions and foreach loops.
    """
    ...


class Enumeration:
    """ Customized Java enumeration.

    Enumerations are used rarely in Java, but can be iterated like a Java
    iterable using Python.
    """
    ...



import jpype


class Thread:
    """ Thread support for Python

    JPype adds  methods to ``java.lang.Thread`` to interact with 
    Python threads.  These methods are all classmethods that 
    act on the current Python thread.
    """

    isAttached = jpype._jthread._JThread.isAttached
    attach = jpype._jthread._JThread.attach
    attachAsDaemon = jpype._jthread._JThread.attachAsDaemon
    detach = jpype._jthread._JThread.detach


class AutoCloseable:
    """ Customizer for ``java.lang.AutoCloseable`` and ``java.io.Closeable``

    This customizer adds support of the ``with`` operator to all Java
    classes that implement the Java ``AutoCloseable`` interface.

    Example:

    .. code-block:: python

        from java.nio.file import Files, Paths
        with Files.newInputStream(Paths.get("foo")) as fd:
          # operate on the input stream

        # Input stream closes at the end of the block.

    """
    ...



# *****************************************************************************
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   See NOTICE file for details.
#
# *****************************************************************************
from __future__ import annotations

import typing

# This is a superset of the keywords in Python.
# We use this so that jpype is a bit more version independent.
# Adding and removing keywords from this list impacts the exposed interfaces,
# and therefore is technically a breaking change.
_KEYWORDS = {
    'False',
    'None',
    'True',
    'and',
    'as',
    'assert',
    'async',
    'await',
    'break',
    'class',
    'continue',
    'def',
    'del',
    'elif',
    'else',
    'except',
    'exec',
    'finally',
    'for',
    'from',
    'global',
    'if',
    'import',
    'in',
    'is',
    'lambda',
    'nonlocal',
    'not',
    'or',
    'pass',
    'print',  # No longer a Python 3 keyword. Kept for backwards compatibility.
    'raise',
    'return',
    'try',
    'while',
    'with',
    'yield',
}


def pysafe(s: str) -> typing.Optional[str]:
    """
    Given an identifier name in Java, return an equivalent identifier name in
    Python that is guaranteed to not collide with the Python grammar.

    """
    if s.startswith("__") and s.endswith("__") and len(s) >= 4:
        # Dunder methods should not be considered safe.
        # (see system defined names in the Python documentation
        # https://docs.python.org/3/reference/lexical_analysis.html#reserved-classes-of-identifiers
        # )
        return None
    if s in _KEYWORDS:
        return s + "_"
    return s



# This is the audit script for checking the GC linker.
#
# It beats to memory to see if everything holds up,  but not suitable for
# the test suite.  It is run in conjuction with the printf statement in
# jp_gc.cpp to see how aggressive we are being about running the gc.
# The goal is to keep the memory as low as possible while still maintaining
# good speed.  It has to be checked with 3 different block sizes to
# verify function (<1k, 10k, >1Mb) as different behaviors occur at different
# usage points.

from os import path
import _jpype
import jpype
from jpype.types import *
import numpy as np
import gc
import time

# print(gc.callbacks)
#
# def callHook(*args):
#    jpype.java.lang.System.gc()
#
# gc.callbacks.append(callHook)

trials = 200000
tally = np.zeros((trials,), dtype=np.int8)


class DestructionTracker:
    del_calls = 0
    init_calls = 0

    def __init__(self, i, obj):
        self.index = i
        self.obj = obj
        DestructionTracker.init_calls += 1
        tally[i] = 1
        super().__init__()

    def __del__(self):
        tally[self.index] += 1
        DestructionTracker.del_calls += 1

    def callback(self, message):
        pass


if __name__ == '__main__':
    jpype.startJVM(classpath=['test/classes', 'project/jpype_java/dist/*'],
                   convertStrings=True)

    print()
    kB = (1024 / 8)
    MB = (1024**2 / 8)

    fixture = JClass("jpype.common.Fixture")()
    for i in range(trials):
        x = np.arange(int(10 * kB), dtype=np.int64)

        interface = jpype.JProxy("java.io.Serializable",
                                 dict={'callback': DestructionTracker(i, x).callback})
        interface_container = fixture.callObject(interface)

        if (i % 1000) == 0:
            stats = _jpype.gcStats()
            print("created=", DestructionTracker.init_calls,
                  "  destroyed=", DestructionTracker.del_calls,
                  "  delta=", DestructionTracker.init_calls - DestructionTracker.del_calls,
                  "  current=", stats['current'],
                  "  min=", stats['min'],
                  "  max=", stats['max'],
                  "  triggered=", stats['triggered'],
                  )
            time.sleep(1)
#        print(_jpype.gcStats())
        del interface, interface_container
#        if DestructionTracker.del_calls != 0:
#            print(f'{i} We have deleted something: {DestructionTracker.del_calls}')
#        else:
#            print(f'{i} Still no deletion on it {i}\n', end="")
    print()
print(np.sum(tally == 2))
jpype.shutdownJVM()
print(np.sum(tally == 2))




class AssertRaises(object):
    def __init__(self, expected):
        self.expected = expected

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is None:
            try:
                exc_name = self.expected.__name__
            except AttributeError:
                exc_name = str(self.expected)
            raise AssertionError("{0} not raised".format(exc_name))
        if not issubclass(exc_type, self.expected):
            # let unexpected exceptions pass through
            return False
        return True



# *****************************************************************************
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   See NOTICE file for details.
#
# *****************************************************************************
import jpype
import common


def callFunctional(cls, i):
    fun = cls @(lambda x: x)
    return fun.call(i)


class FunctionalTestCase(common.JPypeTestCase):
    
    def _checkValidFunctional(self, name, value):
        cls = jpype.JClass("jpype.functional." + name)
        self.assertEqual(callFunctional(cls, value), value)
    
    def testAnnotated(self):
        self._checkValidFunctional("Annotated", 0)

    def testNonAnnotated(self):
        self._checkValidFunctional("NonAnnotated", 1)

    def testExtendsFunctional(self):
        self._checkValidFunctional("ExtendsFunctional", 2)

    def testExtendsNonFunctional(self):
        with self.assertRaises(TypeError):
            self._checkValidFunctional("ExtendsNonFunctional", 3)
    
    def testAnnotatedWithObjectMethods(self):
        self._checkValidFunctional("AnnotatedWithObjectMethods", 4)

    def testNonAnnotatedWithObjectMethods(self):
        self._checkValidFunctional("NonAnnotatedWithObjectMethods", 5)

    def testRedeclaresAnnotated(self):
        self._checkValidFunctional("RedeclaresAnnotated", 6)

    def testRedeclaresNonAnnotated(self):
        self._checkValidFunctional("RedeclaresNonAnnotated", 7)

    def testStaticDefaultMethods(self):
        # Predicate has both static and default methods
        Predicate = jpype.JClass("java.util.function.Predicate")
        fun = Predicate @ (lambda x: True)
        self.assertTrue(fun.test(fun))



from typing import Any, Iterable, Collection, List, Set, Mapping, Tuple, TypeVar, Iterator, Generator, Union, overload, Dict

E = TypeVar('E')
K = TypeVar('K')
V = TypeVar('V')


class _JIterable(Iterable[E]):
    def __iter__(self) -> Iterator[E]: ...


class _JCollection(Collection[E]):
    def __len__(self) -> int: ...

    def __delitem__(self, i: E) -> None: ...

    def __contains__(self, i: Any) -> bool: ...

    def __iter__(self) -> Iterator[E]: ...


class _JList(List[E]):
    pass


class _JMap(Dict[K, V]):
    def __len__(self) -> int: ...

    def __iter__(self) -> Iterator[K]: ...

    def __getitem__(self, ndx: K) -> V: ...


class _JSet(Set[E]):
    pass


class _JMapEntry(Tuple[K, V]):
    pass


class _JIterator(Iterator[E]):
    def __next__(self) -> E: ...


class _JEnumeration(Iterator[E]):
    def __next__(self) -> E: ...



import _jpype
import jpype
from jpype.types import *
import jpype.imports
jpype.startJVM(classpath=['gson-2.8.5.jar', 'gson-2.8.5-javadoc.jar', 'project/jpype_java/dist/*', 'project/jpype_java/jdk-11.0.7_doc-all.zip'])

import org

html = JClass("org.jpype.html.Html")
hw = JClass("org.jpype.html.HtmlWriter")
#jdz = JClass("org.jpype.javadoc.JavadocZip")(Paths.get("project/jpype_java/jdk-8u251-docs-all.zip"))
jde = JClass("org.jpype.javadoc.JavadocExtractor")
#jdf = JClass("org.jpype.javadoc.JavadocTransformer")()
#jdr = JClass("org.jpype.javadoc.JavadocRenderer")()

current = None


def renderClass(cls):
    global current
    jd = jde.getDocumentation(cls)
    #jis = jdz.getInputStream(cls)
    if jd is None:
        return

    print("=========================================================")
    print("CLASS", cls)

    print(jd.description)
    print(jd.ctors)
    print("---------------------------------------------------------")

    if jd.methods is not None:
        for p, v in jd.methods.items():
            print(v)
            print("- - - - - - - - - - - - - - - - - - - - - - - - - - - - -")

    print("---------------------------------------------------------")

    if jd.fields is not None:
        for p, v in jd.fields.items():
            print(v)
            print("- - - - - - - - - - - - - - - - - - - - - - - - - - - - -")
    print("=========================================================")


def renderPackage(pkg):
    for i in dir(pkg):
        print("Test", i)
        try:
            p = getattr(pkg, i)
        except Exception:
            continue
        if isinstance(p, _jpype._JPackage):
            renderPackage(p)
            continue
        if isinstance(p, jpype.JClass):
            renderClass(p)


try:
    renderPackage(jpype.JPackage('com.google.gson'))
except org.jpype.javadoc.JavadocException as ex:
    print("Javadoc Error: ", ex.message())
    print(hw.asString(ex.node))
    raise ex
except Exception as ex:
    print("Error in")
    print(hw.asString(current))
    raise ex



# *****************************************************************************
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   See NOTICE file for details.
#
# *****************************************************************************
import jpype
import subrun
import functools
import os
from pathlib import Path
import unittest

root = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
cp = os.path.join(root, 'classes').replace('\\', '/')


@subrun.TestCase(individual=True)
class StartJVMCase(unittest.TestCase):
    def setUp(self):
        self.jvmpath = jpype.getDefaultJVMPath()

    def testStartup(self):
        with self.assertRaises(OSError):
            jpype.startJVM(convertStrings=False)
            jpype.startJVM(convertStrings=False)

    def testRestart(self):
        with self.assertRaises(OSError):
            jpype.startJVM(convertStrings=False)
            jpype.shutdownJVM()
            jpype.startJVM(convertStrings=False)

    def testInvalidArgsFalse(self):
        with self.assertRaises(RuntimeError):
            jpype.startJVM(
                "-for_sure_InVaLiD",
                ignoreUnrecognized=False, convertStrings=False,
            )

    def testInvalidArgsTrue(self):
        jpype.startJVM(
            "-for_sure_InVaLiD",
            ignoreUnrecognized=True,
            convertStrings=False,
        )

    def testClasspathArgKeyword(self):
        jpype.startJVM(classpath=cp, convertStrings=False)
        assert jpype.JClass('jpype.array.TestArray') is not None

    def testClasspathArgList(self):
        jpype.startJVM(
            classpath=[cp],
            convertStrings=False,
        )
        assert jpype.JClass('jpype.array.TestArray') is not None

    def testClasspathArgListEmpty(self):
        jpype.startJVM(
            classpath=[cp, ''],
            convertStrings=False,
        )
        assert jpype.JClass('jpype.array.TestArray') is not None

    def testClasspathArgDef(self):
        jpype.startJVM('-Djava.class.path=%s' % cp, convertStrings=False)
        assert jpype.JClass('jpype.array.TestArray') is not None

    def testClasspathArgPath(self):
        jpype.startJVM(classpath=Path(cp), convertStrings=False)
        assert jpype.JClass('jpype.array.TestArray') is not None

    def testClasspathArgPathList(self):
        jpype.startJVM(classpath=[Path(cp)], convertStrings=False)
        assert jpype.JClass('jpype.array.TestArray') is not None

    def testClasspathArgGlob(self):
        jpype.startJVM(classpath=os.path.join(cp, '..', 'jar', 'mrjar*'))
        assert jpype.JClass('org.jpype.mrjar.A') is not None

    def testClasspathTwice(self):
        with self.assertRaises(TypeError):
            jpype.startJVM('-Djava.class.path=%s' %
                            cp, classpath=cp, convertStrings=False)

    def testClasspathBadType(self):
        with self.assertRaises(TypeError):
            jpype.startJVM(classpath=1, convertStrings=False)

    def testJVMPathArg_Str(self):
        jpype.startJVM(self.jvmpath, classpath=cp, convertStrings=False)
        assert jpype.JClass('jpype.array.TestArray') is not None

    def testJVMPathArg_None(self):
        # It is allowed to pass None as a JVM path
        jpype.startJVM(
            None,  # type: ignore
            classpath=cp,
        )
        assert jpype.JClass('jpype.array.TestArray') is not None

    def testJVMPathArg_NoArgs(self):
        jpype.startJVM(
            classpath=cp,
        )
        assert jpype.JClass('jpype.array.TestArray') is not None

    def testJVMPathArg_Path(self):
        with self.assertRaises(TypeError):
            jpype.startJVM(
                # Pass a path as the first argument. This isn't supported (this is
                # reflected in the type definition), but the fact that it "works"
                # gives rise to this test.
                Path(self.jvmpath),  # type: ignore
                convertStrings=False,
            )

    def testJVMPathKeyword_str(self):
        jpype.startJVM(
            classpath=cp,
            jvmpath=self.jvmpath,
            convertStrings=False,
        )
        assert jpype.JClass('jpype.array.TestArray') is not None

    def testJVMPathKeyword_Path(self):
        jpype.startJVM(jvmpath=Path(self.jvmpath), classpath=cp, convertStrings=False)
        assert jpype.JClass('jpype.array.TestArray') is not None

    def testPathTwice(self):
        with self.assertRaises(TypeError):
            jpype.startJVM(self.jvmpath, jvmpath=self.jvmpath)

    def testBadKeyword(self):
        with self.assertRaises(TypeError):
            jpype.startJVM(invalid=True)  # type: ignore



# *****************************************************************************
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   See NOTICE file for details.
#
# *****************************************************************************
import keyword  # From the standard library.

import pytest

import jpype
import common


@pytest.mark.parametrize(
    "identifier",
    list(keyword.kwlist) + [
        '__del__',
        # Print is no longer a keyword in Python 3, but still protected to
        # avoid API breaking in JPype v1.
        'print',
        '____',  # Not allowed.
    ]
)
def testPySafe__Keywords(identifier):
    safe = jpype._pykeywords.pysafe(identifier)
    if identifier.startswith("__"):
        assert safe is None
    else:
        assert isinstance(safe, str), f"Fail on keyword {identifier}"
        assert safe.endswith("_")


@pytest.mark.parametrize(
    "identifier",
    [
        'notSpecial',
        '__notSpecial',
        'notSpecial__',
        '_notSpecial_',
        '_not__special_',
        '__', '___',  # Technically these are fine.
    ])
def testPySafe__NotKeywords(identifier):
    safe = jpype._pykeywords.pysafe(identifier)
    assert safe == identifier


class AttributeTestCase(common.JPypeTestCase):
    def testPySafe(self):
        cls = jpype.JPackage("jpype").attr.TestKeywords
        self.assertTrue(hasattr(cls, "__leading_double_underscore"))
        self.assertFalse(hasattr(cls, "__dunder_name__"))
